#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libgen.h>
#include <getopt.h>
#include "wstationd.h"


static char *exec_name;


static void print_help()
{
	fprintf(stdout, "usage: %s [-p port] [-b addr] directory\n", exec_name);
	fprintf(stdout, "  -p, --port <port>     Port to listen on (default: 10800)\n");
	fprintf(stdout, "  -b, --bind <address>  Address to bind to (default: 0.0.0.0)\n");
	fprintf(stdout, "  -h, --help            Print this message and exit\n");
	fprintf(stdout, "  -v, --version         Print version and exit\n");
}


int main(int argc, char *argv[])
{
	const struct option longopts[] = {
		{ "port", required_argument, NULL, 'p' },
		{ "bind", required_argument, NULL, 'b' },
		{ "version", no_argument, NULL, 'v' },
		{ "help", no_argument, NULL, 'h' },
		{ NULL, 0, NULL, 0 }
	};

	int flag = 1;
	int sockfd = 0, connfd = 0;
	int port = 10800;
	uint32_t address = INADDR_ANY;
	struct sockaddr_in sock;

	struct sockaddr_in client;
	socklen_t client_sz = sizeof(client);
	char timebuff[15], fnbuff[35];

	time_t timeraw;
	struct tm* timeptr;

	exec_name = basename(argv[0]);

	for(int ch; (ch = getopt_long(argc, argv, "b:p:vh0", longopts, NULL)) != -1;){
		switch(ch){
			case 'p':
				port = (int)strtol(optarg, NULL, 10);
				if(port < 1 || port > 65535){
					fprintf(stderr, "%s: invalid port number\n", exec_name);
					print_help();
					goto shutdown_error;
				}
			break;

			case 'b':
				if(inet_pton(AF_INET, optarg, &address) == -1){
					fprintf(stderr, "%s: invalid bind address\n", exec_name);
					print_help();
					goto shutdown_error;
				}
			break;

			case 'v':
				fprintf(stdout, "wstationd v%s (%s %s)\n", VERSION,__DATE__, __TIME__);
			goto shutdown_clean;

			case '?':
			case 'h':
				print_help();
			goto shutdown_clean;
		}
	}

	// Accept final argument, the directory
	if(argc == (optind+1)){
		// Fail if we can't chdir to that directory
		if(chdir(argv[argc-1]) == -1){
			fprintf(stderr, "%s: cannot change directory - %s\n",
				exec_name, strerror(errno)
			);
			goto shutdown_error;
		}
	} else {
		print_help();
		goto shutdown_error;
	}
	
	// Establish the socket
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if(sockfd == -1){
		fprintf(stderr, "%s: cannot open socket - %s\n",
			exec_name, strerror(errno)
		);
		goto shutdown_error;
	}

	// This sets REUSE on the socket so it's easily reallocated if
	// this program dies
	setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(flag));

	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_addr.s_addr = address;
	sock.sin_port = htons(port);

	// Bind our socket to the specified address and port
	if(bind(sockfd, (struct sockaddr*)&sock, sizeof(sock)) < 0){
		fprintf(stderr, "%s: cannot bind - %s\n",
			exec_name, strerror(errno)
		);
		goto shutdown_error;
	}

	// Listen on port
	if(listen(sockfd, 5) != 0){
		fprintf(stderr, "%s: cannot listen - %s\n",
			exec_name, strerror(errno)
		);
		goto shutdown_error;
	}

	while(1){
		// Wait for a connection on our bound socket
		connfd = accept(sockfd, (struct sockaddr*)&client, &client_sz);
		if(connfd == -1){
			fprintf(stderr, "%s: accept failed - %s\n",
				exec_name, strerror(errno)
			);
			goto shutdown_error;
		}

		// Collect the time of the connection, process it into 
		// a short string
		time(&timeraw);
		timeptr = localtime(&timeraw);
		strftime(&timebuff[0], 15, "%Y%m%d%H%M%S", timeptr);
		timeptr = NULL;

		// Take the short date/time connection string and add the client address
		// to create a filename
		snprintf(&fnbuff[0], 34, "%s-%s.dat", timebuff, inet_ntoa(client.sin_addr));

		fprintf(stdout, "filename: %s\n", fnbuff);

		// We're done, close the connection
		close(connfd); connfd = 0;
	}

	shutdown_clean:
	if(connfd > 0) close(connfd);
	if(sockfd > 0) close(sockfd);
	return 0;

	shutdown_error:
	if(connfd > 0) close(connfd);
	if(sockfd > 0) close(sockfd);
	return 1;
}
